#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>

#define MAXRB 110
#define NBOFENODEB 50


struct user{
	double x;
	double y;
	double tx;
	double dist;	//Distance between this user and its BS
	int rb;		//If 0 it interfers
};



struct node{
	double x;
	double y;
	double dist;
	int interferer;	//=1 if this eNodeB interfers, 0 otherwise
	double tx;	//tx to a typical user (in the cell of this eNodeB)
	int txSet;	//indicates if tx has been set or not (1 if yes / 0 otherwise)
	/* Only for the hueristic */
	int nbOfUsers; 
	struct user	users[MAXRB]; 
};


int totalUsers; 
double cdfMMRB[MAXRB];	//cdf of the M/M/RB/RB queue
double rho, cons;
double lambdaB=2.25;	//Default value
double distBSUser=0.05;	//Default vaue
int NBOFRB=15;		//Default value
int Delta=3;		//Default value
struct node eNodeB[100];
int strategy=1;		//1:MMRB 2:thinning 3:Delta

double convertWattToDB(double value)
{
	if(value==0)
	{
		fprintf(stderr,"Error convertWattToDB(): the power is <=0.0 (return -200)\n");
		return(-200.0);
	}

	return(10.0*log10(value));
}

double convertWattToDBm(double value)
{
	if(value==0)
	{
		fprintf(stderr,"Error convertWattToDBm(): the power is <=0.0 (return -200)\n");
		return(-200.0);
	}

	return(10.0*log10(value)+30.0);
}


/******** Distribution of M/M/RB/RB *************/

double uniforme(double a, double b)
{
	if(b<a)
	{
		fprintf(stderr,"Warning: in function uniform(), b must be greater than a\n");
	}
    
	return(1.0*(b-a)*rand()/(RAND_MAX+1.0) + a);
}

double factoriel(int i)
{
	int k;
	double res=1.0;

	if(i==0 || i==1) return(1.0);

	for(k=2;k<=i;k++) res*=1.0*k;
	return(res);
}

double calculConstante()
{
	int i, k;
	double res, tmp;
	
	res=1.0+1.0*NBOFRB*rho;//terme0+terme1

	if(NBOFRB<16) 
	{
		for(i=2;i<=NBOFRB;i++) res+=pow(1.0*NBOFRB*rho,i)/factoriel(i);
	} else {
		for(i=2;i<=NBOFRB;i++)
		{ 
			tmp=1.0;
			for(k=1;k<=i;k++) tmp*=(1.0*NBOFRB*rho)/(1.0*k);
			res+=tmp;
		}
	}

	return(1.0/res);
}

double probaMMRBRB(int i)
{	
	int k;
	double res;

	if(i>NBOFRB) return(0.0);
	if(i<0) return(0.0);
	if(i==0) return(cons);
	
	if(i<16) res=pow(1.0*NBOFRB*rho,1.0*i)*cons/factoriel(i);
	else {
		res=1.0;
		for(k=1;k<=i;k++) res*=(1.0*NBOFRB*rho)/(1.0*k);
		res*=cons;
	}

	//printf("res=%e (i=%d)\n",res,i);
	return(res);
}

void initCdfMMRBRB()
{
	int i;
	double tmp;

	cdfMMRB[0]=cons;

	for(i=1;i<=NBOFRB;i++)
		cdfMMRB[i]=cdfMMRB[i-1]+probaMMRBRB(i); 

	//DEBUG: mean Nuber of customers
	static int once=0;
	if(once==0)
	{
		tmp=0.0;
		for(i=0;i<=NBOFRB;i++)
		{
			tmp+=1.0-cdfMMRB[i]; 
		}
		//fprintf(stderr,"Mean number of customer=%f (should be NBOFRB/Delta=%f) \n",tmp,(1.0*NBOFRB)/(1.0*Delta));
		once=1;
	}
}

int simuRVMMRBRB()
{
	int i=0;
	double u, tmp;

	tmp=cdfMMRB[0];
	u=uniforme(0.0,1.0);
	while(i<NBOFRB && u>tmp)
	{ 
		i++;
		tmp=cdfMMRB[i];
	}
	return(i);
}


int generateDiGivenWi()
{
	int cpt=0, sum, D0;
	static int init=0;

	if(init==0) 
	{
		cons=calculConstante(rho);
		initCdfMMRBRB(rho);
	}

	//Di sachant que wi=1
	sum=0;
	while(cpt<10) //we loop 10 times over RB
	{
		while(sum<NBOFRB)
		{
			sum+=simuRVMMRBRB(rho);
		}
		cpt++;
		sum-=NBOFRB;
	}

	//On cherche maintenant le prochain Di (D0 ci dessous) tel que wi=1
	while(sum<NBOFRB)
	{
		D0=simuRVMMRBRB(rho);
		sum+=D0;
	}
	//On a trouve !!!

	return(D0);
}




//i is the eNodeB (ordered w.r.t. the distance to the origin)
int genereDi(int i)
{

	if(i==0) return(generateDiGivenWi());
	if(i>0) return(simuRVMMRBRB());
	
	fprintf(stderr,"Error in genereDi(): i=%d --> abort\n",i);
	exit(1);
}

//Return Wi i.e. 0 or one.
//The argument is usefull only when i=0
int returnWiMMRB(int i)
{
	static int lastUsedRB=0;
	int numberOfUsers, wi=0;

	if(i==0)
	{  
		numberOfUsers=genereDi(i);//Number of user allocated to this eNodeB
		lastUsedRB=numberOfUsers;//Slots are counted from 0 to genereDi()-1
		lastUsedRB-=rand()%lastUsedRB + 1;//We take one slot randomly among genereDi and shift the slots (the slot of the typical user is 0)
	} else {
		numberOfUsers=genereDi(i);//Number of user allocated to this eNodeB
		lastUsedRB+=numberOfUsers;
		if(lastUsedRB>=NBOFRB)
		{
			lastUsedRB=lastUsedRB%NBOFRB;
			wi=1;
		}
	}

	totalUsers+=numberOfUsers;//Used only for the heuristic
	eNodeB[i].nbOfUsers=numberOfUsers;

	//fprintf(stderr,"--DEBUG lastUsedRB=%d\n",lastUsedRB);
	return(wi);
}

int returnWi(int i)
{
	double u;

	if(strategy==1 || strategy==4) return(returnWiMMRB(i));
	if(strategy==2) 
	{
		u=uniforme(0.0,1.0);

		if(u <  (1.0/(1.0*Delta))) return(1);
		else return(0);
	}
	if(strategy==3) 
	{
		if( (i%Delta)==0 ) return(1);
		else return(0);
	}

	fprintf(stderr,"Error in returnWi(): the argument i (=%d) is invalid\n",i);
	exit(1);
}

/********** End of the set of functions used to determine wi ***********************/

//Draw the distance ||X0||
double distanceX0()
{
	double unif, dist;
	
	unif=uniforme(0.0,1.0);
	dist=sqrt(- log(1.0-unif)/(lambdaB*1.25*3.141592));


	return(dist);
}

//u is the distance of the previous ENodeB
double distanceNextENodeB(double u)
{
	double unif, dist;
	
	unif=uniforme(0.0,1.0);
	dist=sqrt(u*u - log(1.0-unif)/(lambdaB*3.141592));


	return(dist);
}

//Distance en km!!!!
double pathLoss(double dist)
{
	double res;

	//On borne a 1m
	if(dist<=0.001) return( 0.0295120 );

	res=-128.1-37.6*log10(dist);
	res=pow(10.0,res/10.0);//Convert dB -> linear
	
	return(res);
}

//Compute the distance between node1 and node2
double distancePoints(struct node node1, struct node node2)
{
	double distance;
	distance=(node1.x-node2.x)*(node1.x-node2.x);
	distance+=(node1.y-node2.y)*(node1.y-node2.y);
    
	return(sqrt(distance));
}//end of distancePoints()

int findClosestENodeB(struct node user, struct node* eNodeB, int nbOfENodeB)
{
	int i, bestENode=-1;
	double distance, best=9999999.0;
	
	for (i=0; i<nbOfENodeB; i++)
	{
	 	distance=distancePoints(user, eNodeB[i]);
		if(distance<best)
		{ 
			best=distance;
			bestENode=i;
		}
	}

	if(bestENode==-1)
	{
		fprintf(stderr,"Error: updateUsers fail to find an eNodeB\n");
		exit(3);
	}

	return(bestENode);
}

//En Watt!!!
double transmissionPower(double dist)
{

	if(dist	< 0.05){ return(0.0476);}
	if(dist	< 0.10){ return(0.0645);}
	if(dist	< 0.15){ return(0.296);}
	if(dist	< 0.20){ return(0.8747);}
	if(dist	< 0.25){ return(2.024);}
	if(dist	< 0.30){ return(4.017);}
	if(dist	< 0.35){ return(7.17);}
	if(dist	< 0.40){ return(11.85);}
	if(dist	< 0.45){ return(18.45);}
	return(27.42);//>0.45

}

//target is the number of interferer and maxDist is the farthest eNodeB
void setTxUser(int target, double maxDist)
{
	int j=0, eNodeBID, borne=30;
	double d;
	struct node point;


	d=maxDist*2.4;
	while( j<target )
	{
        	point.x = d*((1.0 * rand())/(RAND_MAX+1.0)) - (d/(2.0));
        	point.y = d*((1.0 * rand())/(RAND_MAX+1.0))  - (d/(2.0));
		eNodeBID=findClosestENodeB(point, eNodeB, NBOFENODEB);

		if( eNodeBID!=0 && (eNodeB[eNodeBID].interferer==1) && (eNodeB[eNodeBID].txSet==0) )//We find a user for this interferer
		{
			j++;
			eNodeB[eNodeBID].txSet=1;
			eNodeB[eNodeBID].tx=transmissionPower(distancePoints(point,eNodeB[eNodeBID]));
		}
	}
}

//strategy 4: associate users to an eNodeB
void setUserForAnENodeB(int eNodeBid)
{
	int j=0, closestENodeB, numberOfUsers=0;
	double d;
	struct node point, zero;

	zero.x=0.0;
	zero.y=0.0;


	d=5.0/sqrt(3.14159*lambdaB);
	while(numberOfUsers<eNodeB[eNodeBid].nbOfUsers)
	{

		//Around eNodeBid
        	point.x = eNodeB[eNodeBid].x + d*( 0.5 - (1.0 * rand())/(RAND_MAX+1.0) );
        	point.y = eNodeB[eNodeBid].y + d*( 0.5 - (1.0 * rand())/(RAND_MAX+1.0) );

		closestENodeB=findClosestENodeB(point, eNodeB, NBOFENODEB);

		if( closestENodeB == eNodeBid )//We find a user for this interferer
		{

			eNodeB[eNodeBid].users[numberOfUsers].x=point.x;
			eNodeB[eNodeBid].users[numberOfUsers].y=point.y;
			eNodeB[eNodeBid].users[numberOfUsers].tx=transmissionPower(distancePoints(point,eNodeB[eNodeBid]));
			eNodeB[eNodeBid].users[numberOfUsers].dist=distancePoints(point,eNodeB[eNodeBid]);
			eNodeB[eNodeBid].users[numberOfUsers].rb=-1;
			numberOfUsers++;
		}
	}
}

double interferenceRB(int eNode, int userId, int rb)
{
	int i, j;
	double interference=0.0;
	struct node thisUser;

	for(i=0;i<30;i++)
	{ 
		if(i!=eNode)
		{
			for(j=0;j<eNodeB[i].nbOfUsers;j++)
			{ 
				if(eNodeB[i].users[j].rb==rb)
				{ 
					thisUser.x=eNodeB[eNode].users[userId].x;
					thisUser.y=eNodeB[eNode].users[userId].y;

					interference+=eNodeB[i].users[j].tx*pathLoss(distancePoints(eNodeB[i], thisUser));
				}
			}
		}
	}

	return(interference);
}

//Tester non deja alloue
int isAllocated(int eNode, int rb)
{
	int j; 

	for(j=0;j<eNodeB[eNode].nbOfUsers;j++)
	{ 
		if(eNodeB[eNode].users[j].rb==rb) return(1);
	}

	return(0);
}



void heuristicAllocation()
{
	int i=0, j, restant, next, userId, eNode, bestRB, tmp;
	int tabUser[5000];	
	double bestInterf, interf;

	if(totalUsers >=5000)
	{
		fprintf(stderr,"Error heuristicAllocation(): array tabUser is not big enough (%d for %d users)\n",5000,totalUsers);
		exit(1);
	}

	for(i=0;i<totalUsers;i++) tabUser[i]=0;//0 if done, 1 otherwise

	//Prendre un noeud au hasard ... utiliser un done et avancer ... sur le nombre d'utilisateurs restant
	i=0;
	while(i < totalUsers)
	{
		restant=totalUsers-i;	
		next = rand()%restant;//au hasard entre 0 et restant-1
		//fprintf(stderr,"--Heuristic allocation i=%d sur %d  restant=%d next=%d\n",i,totalUsers,restant,next);
		
		j=0; userId=0;
		while(j<next+1)
		{
			//fprintf(stderr,"--loop 1 -- userId=%d j=%d\n",userId,j);
			if(tabUser[userId]==0) j++;
			userId++;
		}
		if(userId>0) userId--;
		tabUser[userId]=1;
		//fprintf(stderr,"-- userId=%d j=%d\n",userId,j);

		//Slot allocation for the user userId

		//What is its eNodeB?
		j=0;eNode=0;
		while(j<=userId)
		{ 
			//Il faut ajouter le nombre de noeuds restant à 0 
			j+=eNodeB[eNode].nbOfUsers; 
			eNode++;
		}

		if(eNode>0) eNode--; 
		
		if(eNode>0) userId=userId-j+eNodeB[eNode].nbOfUsers;
		//fprintf(stderr,"--DEBUG-- totalUser=%d restant=%d pick=%d eNodeB=%d eNodeB.nbOfUsers=%d\n",totalUsers,restant,userId,eNode,eNodeB[eNode].nbOfUsers);
			
		//Allocation RB	
		bestInterf=999999.99; bestRB=-1;
		for(j=0;j<NBOFRB;j++)
		{
			//teste si non deja alloue a cet eNodeB
			if(isAllocated(eNode,j)==0) 
			{
				interf=interferenceRB(eNode, userId, j);
				//fprintf(stderr,"Compute interf -- rb=%d interf=%e bestInterf=%e\n",j,interf,bestInterf);
				if(interf<bestInterf)
				{
					bestInterf=interf;
					bestRB=j;
				}
			}
		}	

		//fprintf(stderr,"Allocation -- eNode=%d user=%d over %d rb=%d bestInterf=%e\n",eNode,userId,eNodeB[eNode].nbOfUsers,bestRB,bestInterf);
		if(bestRB!=-1) 
		{
			eNodeB[eNode].users[userId].rb=bestRB;
			eNodeB[eNode].tx=eNodeB[eNode].users[userId].tx;//sert a quelquechose???? En plus (a la fin) lie a un des utilisateurs mais lequel?
		}

		i++;
	}


}


void generateENodeB()
{
	int i, target=0;
	double u=0.0, angle;

	for(i=0;i<NBOFENODEB;i++)
	{ 
		if(i==0)
		{ 
			if(distBSUser<=0.0) u=distanceX0();//If the distance X0 needs to be drawn
			if(distBSUser>0.0) u=distBSUser;//If the distance X0 is set 
		} else u=distanceNextENodeB(u);
		eNodeB[i].dist=u;

		angle=uniforme(0.0,2.0*3.141596);
		eNodeB[i].x=u*cos(angle);
		eNodeB[i].y=u*sin(angle);
		
		eNodeB[i].txSet=0;

		//We generated NBOFENODEB eNodeB to avoid edge effect but we consider only interferers in the 30 first eNodeB except if we have to plot it
		if(i==0 && (strategy==1 || strategy==4))	eNodeB[i].interferer=returnWi(0);

		if( i>0 && i<30) 
		{
			eNodeB[i].interferer=returnWi(i);
			if(eNodeB[i].interferer==1)
			{ 
					target++;
			}

		} else eNodeB[i].interferer=0;
	}

	//Strategy 4: we set the users and allocation
	if(strategy==4)
	{

		for(i=0;i<30;i++) setUserForAnENodeB(i);
		heuristicAllocation();

	}
	else setTxUser(target,u);

}


//Compute interference
double interference()
{
	int i;
	double interference=0.0;

	generateENodeB();

	for(i=0;i<30;i++)
	{ 
		if(eNodeB[i].interferer==1)
		{ 
			interference+=eNodeB[i].tx*pathLoss(eNodeB[i].dist);
		}
	}

	return(interference);
}

//This function is not called, but you can call it if you want to debug or print allocation for a given sample. 
void printAllocation()
{
	int i,j;

	for(i=0;i<30;i++)
	{
		fprintf(stderr,"-- Allocation eNode %d --\n",i);
		fprintf(stderr,"- eNode %d has %d users\n",i,eNodeB[i].nbOfUsers);
	
		for(j=0;j<eNodeB[i].nbOfUsers;j++) fprintf(stderr,"- users %d uses rb=%d\n",j,eNodeB[i].users[j].rb);
	}
}

void generateSample(int nbOfSamples)
{
	int i, rbTypical;
	double interf, sinr, Noise=0.0;


	for(i=0;i<nbOfSamples;i++)
	{	
		if(i%1000==0) fprintf(stderr,"Ech %d sur %d\n",i,nbOfSamples);

		if(strategy<4) interf=interference();
		else
		{
			totalUsers=0;//Will be updated in returnWiMMRB (only)
			generateENodeB();


			//Set interferers - we consider a node bound at eNodeB 0 
			rbTypical=eNodeB[0].users[0].rb;

			if(rbTypical<0)
			{
				fprintf(stderr,"Error heuristicAllocation(): there is no node attached to eNodeB 0\n");
				exit(2);
			}

			interf=interferenceRB(0, 0, eNodeB[0].users[0].rb);

			if((interf+Noise)>0.0) sinr=pathLoss(eNodeB[0].users[0].dist)*eNodeB[0].users[0].tx/(interf+Noise);
			else {
				fprintf(stderr,"Warning strategy 4 generateSample(): interf=0.0\n");
			}

		}


		if(strategy!=4) sinr=pathLoss(eNodeB[0].dist) * transmissionPower(eNodeB[0].dist) / (interf+Noise);
		

		//print samples
		fprintf(stderr,"%e %e\n",interf,sinr);
	}

}



int main(int argc, char* argv[])
{
	int i, nbOfSamples=1;

	srand(time(NULL));

	if(argc>1 && strcmp(argv[1],"-h")==0)
	{
		fprintf(stderr,"Usage:%s strategy Delta NBOFRB nbOfSamples distUserBS(in meters) intensity\n",argv[0]);
		fprintf(stderr,"strategy: 1-MMRB 2-thinning 3-Delta 4-heuristic\n");
		fprintf(stderr,"distUserBS: if <=0.0 this distance is drawn randomly (under Palm) otherwise it is set to the given distance.\n");
		exit(0);
	}

	if(argc>1) strategy=atoi(argv[1]);
	if(argc>2) Delta=atoi(argv[2]);
	if(argc>3) NBOFRB=atoi(argv[3]);
	if(argc>4) nbOfSamples=atoi(argv[4]);
	if(argc>5) distBSUser=atof(argv[5])/1000.0;
	if(argc>6) lambdaB=atof(argv[6]);



	rho=1.0/(1.0*Delta);

	generateSample(nbOfSamples);

	return(0);

}





